{% for name, params in interfaces.items() %}
    {% with name=name, params=params, common={"base_platform_type": base_platform_type, "cdp": context.cdp, "lldp": context.lldp} %}
        {% include "interface/base_interface.jinja2" %}
    {% endwith %}
!
{% endfor %}