! Generated Template Config
!
!
no service pad
service tcp-keepalives-in
service tcp-keepalives-out
service tcp-keepalives-in
service tcp-keepalives-out
!
hostname {{ context.hostname }}
!
!
clock timezone CET 1 0
clock summer-time CEST recurring last Sun Mar 2:00 last Sun Oct 2:00
!
!
{% if context.domain_name %}
ip domain-name {{ context.domain_name }}
{% endif %}
{% for name_server in context.name_servers  %}
ip name-server {{ name_server }}
{% endfor %}
!
{% if "vtp" in context.keys() %}
vtp mode {{ context.vtp.mode|default("transparent", true) }}
{% endif %}
!
{% if not context.cdp %}
no cdp run
{% endif %}
!
!
!
{% if "vrfs" in context.keys() %}
{% include "vrfs.jinja2" %}
{% endif %}
!
!
!
{% include "vlans.jinja2" %}
!
!
!
{% with interfaces=context.interfaces, base_platform_type=context.base_platform_type %}
{% include "interface/interfaces.jinja2" %}
{% endwith %}
!
!
{% if "routing" in context.keys() %}
{% include "routing/static_routes.jinja2" %}
{% endif %}
